package client;

import java.rmi.RemoteException;
import java.util.Scanner;

import transaction.ITravelAgency;
import transaction.InvalidTransactionException;
import transaction.ResourceType;
import transaction.TransactionAbortedException;
import transaction.TravelAgencyImpl;

/**
 * A console to operate on a resource manager. Use this
 * console to simulate single threaded access to a
 * resource manager. The console accepts the following
 * command syntax:
 * 
 * <ul>
 * <li>quit console: q
 * <li>add resource: ar <resource_type ('h'|'f'|'c')> <rid> <num> <price>
 * <li>update resource: ur <resource_type ('h'|'f'|'c')> <rid> <num> <price>
 * <li>add customer: nc <resource_type ('h'|'f'|'c')> <custName>
 * <li>reserve resource: res <resource_type ('h'|'f'|'c')> <rid> <custName>
 * <li>reserve itinerary: ri <rid_hotel> <rid_flight> <rid_car> <custName>
 * <li>query bill: bill <custName> 
 * <li>query availability: av <resource_type ('h'|'f'|'c')> <rid>
 * <li>query price: pr <resource_type ('h'|'f'|'c')> <rid>
 * <li>commit
 * <li>abort
 * <li>dieBeforePrepare: dbp <resource_type ('h'|'f'|'c')>
 * <li>dieAfterPrepare: dap <resource_type ('h'|'f'|'c')>
 * <li>dieBeforeCollecting: dbc
 * <li>dieAfterCollecting: dac
 * 
 * 
 * </ul>
 * @author bross
 *
 */
public class TravelAgencyTerminal {
	
	ITravelAgency ta = new TravelAgencyImpl();		
	
	private void run(){
		String command;
		String[] splitted;
		Scanner in = new Scanner(System.in);
		while(true){
			System.out.print("> ");
			command = in.nextLine();
			
			if(command.length() == 0)
				continue;
			
			splitted = command.split("\\s");
			try {				
				
				if(splitted[0].equals("q")){										
					ta.abort();
					return;
				}
								
				else if(splitted[0].equals("ar")){					
					if(splitted.length != 5){
						System.err.println("wrong number of arguments, syntax is: ar resource_type rid num price");
						continue;
					}
					
					ResourceType rt = parseType(splitted[1]);
					if(rt == null){
						System.out.println("unknown resource type");
						continue;
					}
					
					boolean success = ta.addResource(rt, splitted[2],
							Integer.parseInt(splitted[3]),
							Integer.parseInt(splitted[4]));
					
					if(success)
						System.out.println("added resource successfully");
					else
						System.out.println("resource already exists");
				}				
				
				else if(splitted[0].equals("ur")){
					if(splitted.length != 5){
						System.err.println("wrong number of arguments, syntax is: ur resource_type rid num price");
						continue;
					}
					
					ResourceType rt = parseType(splitted[1]);
					if(rt == null){
						System.out.println("unknown resource type");
						continue;
					}
					
					boolean success = ta.updateResource(rt, splitted[2],
							Integer.parseInt(splitted[3]),
							Integer.parseInt(splitted[4]));
					
					if(success)
						System.out.println("updated resource successfully");
					else
						System.out.println("update unsuccessful, probably too many reservations...");
				}
				else if(splitted[0].equals("nc")){				
					if(splitted.length != 3){
						System.err.println("wrong number of arguments, syntax is: nc resource_type custName");
						continue;
					}
					
					ResourceType rt = parseType(splitted[1]);
					if(rt == null){
						System.out.println("unknown resource type");
						continue;
					}
					
					boolean success = ta.newCustomer(rt, splitted[2]);
					
					if(success)
						System.out.println("added customer successfully");
					else
						System.out.println("insert unsuccessfull, maybe customer exists already");
				}
				else if(splitted[0].equals("res")){					
					if(splitted.length != 4){
						System.err.println("wrong number of arguments, syntax is: res resource_type rid custName");
						continue;
					}
					
					ResourceType rt = parseType(splitted[1]);
					if(rt == null){
						System.out.println("unknown resource type");
						continue;
					}
					
					boolean success = ta.reserve(rt, splitted[2], splitted[3]);
					if(success)
						System.out.println("reservation successfull");
					else
						System.out.println("reservation unsuccessfull, maybe customer or resource do not exist or resource not available");
				}
				else if(splitted[0].equals("ri")){					
					if(splitted.length != 5){
						System.err.println("wrong number of arguments, syntax is: ri <rid_hotel> <rid_flight> <rid_car> <custName>");
						continue;
					}
					
					boolean success = ta.reserveItinerary(splitted[1], splitted[2], splitted[3], splitted[4]);
					if(success)
						System.out.println("reservation successfull");
					else
						System.out.println("reservation unsuccessfull, maybe customer or resource do not exist or resource not available");
				}
				else if(splitted[0].equals("bill")){
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: bill custName");
						continue;
					}
					
					int bill = ta.queryBill(splitted[1]);
					if(bill >= 0)
						System.out.println("current bill is " + bill);
					else
						System.out.println("unknown customer");
				}
				else if(splitted[0].equals("av")){
					if(splitted.length != 3){
						System.err.println("wrong number of arguments, syntax is: av resource_type rid");
						continue;
					}
					
					ResourceType rt = parseType(splitted[1]);
					if(rt == null){
						System.out.println("unknown resource type");
						continue;
					}
					
					int av = ta.queryAvailability(rt, splitted[2]);
					if(av >= 0)
						System.out.println("current availability is: " + av);
					else
						System.out.println("unknown resource");
				}
				else if(splitted[0].equals("pr")){
					if(splitted.length != 3){
						System.err.println("wrong number of arguments, syntax is: pr resource_type rid");
						continue;
					}
					
					ResourceType rt = parseType(splitted[1]);
					if(rt == null){
						System.out.println("unknown resource type");
						continue;
					}
					
					int price = ta.queryPrice(rt, splitted[2]);
					if(price >= 0)
						System.out.println("current price is: " + price);
					else
						System.out.println("unknown resource");
				}
				else if(splitted[0].equals("commit")){
					boolean c = ta.commit();
					if(c){
						System.out.println("transaction successfully committed");
					}
					else
						System.out.println("transaction NOT committed");
				}
				else if(splitted[0].equals("abort")){					
					ta.abort();
					System.out.println("transaction successfully aborted");										
				}	
				
				else if(splitted[0].equals("dbp")){
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: dbp resource_type");
						continue;
					}
					
					ResourceType type = parseType(splitted[1]);
					try {
						ta.dieBeforePrepare(type);
					} catch (RemoteException e) {
						e.printStackTrace();
					}
				}
				
				else if(splitted[0].equals("dap")){
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: dbp resource_type");
						continue;
					}
					
					ResourceType type = parseType(splitted[1]);
					try {
						ta.dieAfterPrepare(type);
					} catch (RemoteException e) {
						e.printStackTrace();
					}
				}
				
				else if(splitted[0].equals("dbc")){					
									
					try {
						ta.dieBeforeCollected();
					} catch (RemoteException e) {
						e.printStackTrace();
					}
				}
				
				else if(splitted[0].equals("dac")){					
					
					try {
						ta.dieAfterCollected();
					} catch (RemoteException e) {
						e.printStackTrace();
					}
				}
				
				
				else{
					System.err.println("unknown command");
				}
			} catch (NumberFormatException e) {
				System.err.println("invalid argument: " + e.getMessage());			
			} catch (TransactionAbortedException e) {
				System.err.println("transaction has been aborted by rm");
			} catch (InvalidTransactionException e) {
				//must not happen
				throw new RuntimeException(e);
			}
		}
	}
	
	private ResourceType parseType(String type){
		if("h".equalsIgnoreCase(type))
			return ResourceType.HOTEL;
		if("f".equalsIgnoreCase(type))
			return ResourceType.FLIGHT;
		if("c".equalsIgnoreCase(type))
			return ResourceType.CAR;
		
		return null;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		TravelAgencyTerminal rc = new TravelAgencyTerminal();
		rc.run();
	}

}
