package lockmgr;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class CopyOfLockManager implements ILockManager{
	
	private Hashtable<String, ReentrantReadWriteLock> lockTable;
	private Hashtable<Integer, Set<Lock>> lockedObjects;
	private Hashtable<String, Set<Integer>> lockHolders;
	
	private WaitingGraph wg;
	
	private final Lock lockDataStructure1 = new ReentrantLock(true);
	
	
	
	public CopyOfLockManager() {
		this.lockTable = new Hashtable<String, ReentrantReadWriteLock>();
		this.lockedObjects = new Hashtable<Integer, Set<Lock>>();
		this.lockHolders = new Hashtable<String, Set<Integer>>();
		
		this.wg = new WaitingGraph();
	}

	public boolean lock(int xid, String strData, LockType lockType)
	throws DeadlockException {
		ReentrantReadWriteLock lock = null;
		Set<Integer> lh = null;
		
		//****** ADJUST DATA STRUCTURES ********
		try{
			lockDataStructure1.lock();
			//check for conflict
			lock = lockTable.get(strData);
			if(lock == null){
				//no conflict since no lock exists at all
				//add lock entry
				lock = new ReentrantReadWriteLock();								
				lockTable.put(strData, lock);																												
			}
			
			lh = lockHolders.get(strData);
			if(lh == null){
				lh = Collections.synchronizedSet(new HashSet<Integer>());
				lockHolders.put(strData, lh);
			}
		}
		finally{
			lockDataStructure1.unlock();
		}
		
		//remember locks held by transaction
		//hashtable is already synchronized and we assume that each thread is
		//associated with a unique xid --> no explicit synchronization necessary
		Set<Lock> lo = lockedObjects.get(xid);
		if(lo == null){
			lo = Collections.synchronizedSet(new HashSet<Lock>());	
			lockedObjects.put(xid, lo);
		}
		
		//****** TRY TO ACQUIRE LOCKS ********
		switch(lockType){
		
		//READ
		case READ:
//			System.out.println(Thread.currentThread().getName() + " wants rl(" + strData + ")");
			if(lock.getReadHoldCount() == 0 && lock.getWriteHoldCount() == 0){	
					
				//remember that this transaction wants to acquire a lock
				lh.add(xid);
				if(!lock.readLock().tryLock()){
					//lock could not be acquired --> deadlock detection
					if(deadlock(xid, lh))
						throw new DeadlockException(xid, "rl(" + strData + ")");
										
					lock.readLock().lock();
				}
				
				//read lock has been acquired				
				lo.add(lock.readLock());
				
			}
			break;
			
		//WRITE
		case WRITE:
//			System.out.println(Thread.currentThread().getName() + " wants wl(" + strData + ")");
				
			if(lock.getWriteHoldCount() == 0){
				//lock upgrade?
				if(lock.getReadHoldCount() == 0){
					//NO UPGRADE
					
					if(!lock.writeLock().tryLock()){
						System.out.println("Transaction " + xid + " checking for deadlock");
						if(deadlock(xid, lh)){						
							throw new DeadlockException(xid, "wl(" + strData + ")");
						}					
						System.out.println("Transaction " + xid + " acquiring lock");					
						lock.writeLock().lock();
						System.out.println("Transaction " + xid + " lock acquired");
					}
					lh.add(xid);
					lo.add(lock.writeLock());
									
				}
				else{								
					
					//LOCK UPGRADE							
					//lock upgrade must be atomic, otherwise this thread unlocks and then another thread
					//might acquire the write lock --> this would violate the "no lock after unlock property"!
	//				System.out.println("before synch: " + Thread.currentThread().getName());
										
					
	//					System.out.println("entered synch: " + Thread.currentThread().getName());
					
					if(!lock.writeLock().tryLock()){
						if(deadlock(xid, lh)){
							throw new DeadlockException(xid, "wl(" + strData + ")");
						}
						synchronized (lock.writeLock()) {
							lock.readLock().unlock();
							lo.remove(lock.readLock());
							lock.writeLock().lock();							
						}
						
					}							
					lh.add(xid);
					lo.add(lock.writeLock());
	//					System.out.println("lock upgrade on object " + strData);
					
	//				System.out.println("left synch: " + Thread.currentThread().getName());
				}
			}			
		}				

		return true;
	}
	
	private boolean deadlock(int xid, Set<Integer> lockHolders){
		//add new edge from this transaction to the transaction
		//holding a read/write lock on the object
		
		//a transaction that upgrades its lock from read to write
		//is NOT waiting for itself! --> exclude it from lockHoldes
		Set<Integer> lockHoldersCopy = new HashSet<Integer>();
		for (Integer integer : lockHolders) {
			if(integer != xid)
				lockHoldersCopy.add(integer);
		}
		synchronized(wg){
			if(!wg.addEdges(xid, lockHoldersCopy)){
				return true;
			}
		}
		
		
		return false;
	}
	
	public boolean unlockAll(int xid) {
		
//		System.out.println("calling unlock for " + Thread.currentThread().getName());
		Set<Lock> locks = lockedObjects.get(xid);
		if(locks != null){
//			System.out.println("found " + locks.size() + " locks for thread " + Thread.currentThread().getName());
			for (Lock lock : locks) {				
				try {
					lock.unlock();					
//					System.out.println("unlocked: " + lock);
				} catch (IllegalMonitorStateException e) {
//					System.out.println(lock.toString());
//					e.printStackTrace();
					throw new RuntimeException(e);
				}
			}
			lockedObjects.remove(xid);
//			Iterator<Entry<String, Set<Integer>>> it = lockHolders.entrySet().iterator();
//			Entry<String, Set<Integer>> entry = null;
//			while(it.hasNext()){
//				entry = it.next();
//				entry.getValue().remove(xid);
//				if(entry.getValue().isEmpty())
//					it.remove();
//			}
			for (Set<Integer> lh : lockHolders.values()) {
				lh.remove(xid);
			}
			wg.removeTransaction(xid);
		}
		
		return true;
	}
	
	/**
	 * A simple graph data structure that allows to check for cycles.
	 * @author bross
	 *
	 */
	private class WaitingGraph{
		
		private final Hashtable<Integer, Set<Integer>> adjacencyList =
			new Hashtable<Integer, Set<Integer>>();				

		public synchronized boolean addEdges(int source, Set<Integer> targets){
			
			Set<Integer> set = adjacencyList.get(source);
			if(set == null){
				adjacencyList.put(source,targets);
			}
			else
				set.addAll(targets);
			
			return !containsCycle();
		}
		
		public synchronized void removeTransaction(int xid){
			adjacencyList.remove(xid);
			for (Set<Integer> adjacentVertices : adjacencyList.values()) {
				adjacentVertices.remove(xid);
			}
		}
		
		private synchronized boolean containsCycle(){
			
			//depth first search			
			HashMap<Integer, Integer> status = new HashMap<Integer, Integer>();
						
			for (Integer integer : adjacencyList.keySet()) {
				if(!status.containsKey(integer)){
					try {
						DFSVisit(integer, status);
					} catch (Exception e) {
						return true;
					}					
				}
											
			}
			return false;
		}
		
		private void DFSVisit(int v, HashMap<Integer, Integer> status) throws Exception{
			status.put(v, 1);
			if(adjacencyList.get(v) != null){										
				for (Integer adjacent : adjacencyList.get(v)) {
					if(!status.containsKey(adjacent))
						DFSVisit(adjacent, status);
					else if(status.get(adjacent) == 1) //detect cycle
						throw new Exception();
				}
			}
			
			status.put(v, 2);
		}
	}
}
