package transaction;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import transaction.LogEntry.LogType;

public class ResourceManagerImpl implements IResourceManager {

	protected String rmiName;

	protected Hashtable<Integer, TransactionStatus> localTransactions = new Hashtable<Integer, TransactionStatus>();

	protected Hashtable<Integer, TransactionStatus> distributedTransactions = new Hashtable<Integer, TransactionStatus>();

	protected final AtomicInteger xidGenerator = new AtomicInteger(0);

	protected Hashtable<String, ResourceData> resources = new Hashtable<String, ResourceData>();

	protected Hashtable<String, Set<String>> reservations = new Hashtable<String, Set<String>>();

	protected LogManager logManager;

	protected ITransactionManager tm;

	protected volatile boolean shutdown = false;

	protected File workingdir;

	protected int fileSwitch = 0;

	protected File rmState;
	protected File db0;
	protected File db1;

	protected boolean dieBeforePointerSwitch = false;
	protected boolean dieAfterPointerSwitch = false;
	
	public ResourceManagerImpl() throws IOException {
		
	}

	@Override
	public void start(String workingdir, String rmiName) throws RemoteException {

		this.workingdir = new File(workingdir);
		if (!this.workingdir.isDirectory())
			throw new RuntimeException("working dir is not a directory: "
					+ workingdir);

		this.rmState = new File(workingdir, "rmState.bin");
		this.db0 = new File(workingdir, "db0.dat");
		this.db1 = new File(workingdir, "db1.dat");			

		this.rmiName = rmiName;

		try {
			this.logManager = new LogManager(workingdir);
			readState();
			readDatabase();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		connectTM(ITransactionManager.RMI_TM_NAME);

	}

	@Override
	public boolean connectTM(String rmiName) throws RemoteException {
		Registry registry = LocateRegistry.getRegistry();
		try {
			tm = (ITransactionManager) registry.lookup(rmiName);
		} catch (NotBoundException e) {
			throw new RemoteException("transaction manager could not be bound",
					e);
		}

		return true;
	}

	@Override
	public void abort(int xid) throws InvalidTransactionException {

		TransactionStatus status = distributedTransactions.get(xid);
		if (status == null)
			throw new InvalidTransactionException(xid, "unknown transaction");

		synchronized (localTransactions) {
			localTransactions.put(xid, TransactionStatus.aborted);
			localTransactions.notifyAll();
		}

		logManager.deleteLogsForTransaction(xid);

	}

	@Override
	public int begin() throws ShutdownException {
		if (shutdown)
			throw new ShutdownException();

		int xid = xidGenerator.incrementAndGet();
		localTransactions.put(xid, TransactionStatus.running);
		return xid;
		// throw new
		// UnsupportedOperationException("no local transactions allowed any more");
	}

	@Override
	public boolean beginDistributed(int globalXID) throws ShutdownException,
			RemoteException, InvalidTransactionException {
		if (shutdown)
			throw new ShutdownException();

		if (distributedTransactions.contains(globalXID))
			throw new InvalidTransactionException(globalXID,
					"distributed transaction has already been started: "
							+ globalXID);

		tm.enlist(this.rmiName, globalXID);
		distributedTransactions.put(globalXID, TransactionStatus.running);

		return true;
	}

	@Override
	public boolean prepare(int globalXID) throws InvalidTransactionException,
			RemoteException {

		

		return true;
	}

	@Override
	public boolean commit(int xid) throws TransactionAbortedException,
			InvalidTransactionException {

		validityCheck(xid);

		synchronized (resources) {
			updateDatabase(xid);
			try {
				writeShadow();
				if (dieBeforePointerSwitch)
					System.exit(-1);
				switchPointer();
				if (dieAfterPointerSwitch)
					System.exit(-1);
			} catch (IOException e) {
				e.printStackTrace();

				throw new TransactionAbortedException(xid,
						"error while updating db on disk");
			} finally {
				logManager.deleteLogsForTransaction(xid);
				synchronized (localTransactions) {
					localTransactions.remove(xid);
					localTransactions.notifyAll();
				}

			}
		}

		return true;
	}
	
	

	@Override
	public void restart() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public boolean shutdown() {
		shutdown = true;

		synchronized (localTransactions) {
			while (runningTransactions()) {
				try {
					System.out
							.println("still need to wait for running transactions before shutting down...");
					localTransactions.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
					return false;
				}
			}
		}

		return true;
	}

	private boolean runningTransactions() {
		synchronized (localTransactions) {
			for (TransactionStatus ts : localTransactions.values()) {
				if (ts == TransactionStatus.running) {
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Add a new customer to database. Should return success if customer already
	 * exists.
	 * 
	 * @param xid
	 *            id of transaction.
	 * @param custName
	 *            name of customer.
	 * @return true on success, false on failure.
	 * 
	 * @throws RemoteException
	 *             on communications failure.
	 * @throws TransactionAbortedException
	 *             if transaction was aborted.
	 * @throws InvalidTransactionException
	 *             if transaction id is invalid.
	 */
	public boolean newCustomer(int xid, String custName)
			throws TransactionAbortedException, InvalidTransactionException {

		validityCheck(xid);
		if (!logManager.insertedCustomer(xid, custName)) {
			logManager.log(xid, custName, LogType.INSERT_CUSTOMER, null, null);
			return true;
		} else
			return false;

	}

	/* Return the total price of all reservations held for a customer. */
	public int queryCustomerBill(int xid, String custName)
			throws TransactionAbortedException, InvalidTransactionException {

		validityCheck(xid);

		int bill = 0;
		ResourceData data = null;
		Set<String> customerReservations = getLatestReservations(xid, custName);

		if (customerReservations.size() == 0
				&& !reservations.containsKey(custName))
			return -1;

		for (String reservation : customerReservations) {
			data = getLatestResourceData(xid, reservation);
			if (data != null)
				bill += data.price;
		}

		return bill;

	}

	public boolean addResource(int xid, String rid, int num, int price)
			throws TransactionAbortedException, InvalidTransactionException {

		if (num < 1 || price < 0)
			return false;

		validityCheck(xid);

		ResourceData data = resources.get(rid);
		if (data == null && logManager.insertedResource(xid, rid) == null) {
			data = new ResourceData(num, 0, price);
			logManager.log(xid, rid, LogType.INSERT_RESOURCE, data, null);
			return true;
		} else {
			return false;
		}

	}

	public boolean updateResource(int xid, String rid, int num, int price)
			throws TransactionAbortedException, InvalidTransactionException {

		validityCheck(xid);

		ResourceData data = getLatestResourceData(xid, rid);
		if (data == null)
			return false;

		if (-num > data.totalAvailable - data.reserved)
			return false;

		logManager.log(xid, rid, LogType.UPDATE_RESOURCE, new ResourceData(
				data.totalAvailable + num, data.reserved, price), null);

		return true;
	}

	public int queryResourceAvailability(int xid, String rid)
			throws TransactionAbortedException, InvalidTransactionException {

		validityCheck(xid);

		ResourceData data = getLatestResourceData(xid, rid);

		if (data == null)
			return -1;

		return data.totalAvailable - data.reserved;

	}

	public int queryResourcePrice(int xid, String rid)
			throws TransactionAbortedException, InvalidTransactionException {

		validityCheck(xid);

		ResourceData data = getLatestResourceData(xid, rid);

		if (data == null)
			return -1;

		return data.price;

	}

	public boolean reserveResource(int xid, String rid, String custName)
			throws TransactionAbortedException, InvalidTransactionException {

		if (rid == null)
			return false;

		validityCheck(xid);

		ResourceData data = getLatestResourceData(xid, rid);

		if (data == null || data.totalAvailable - data.reserved == 0)
			return false;

		Set<String> customerReservations = reservations.get(custName);
		if (customerReservations == null
				&& !logManager.insertedCustomer(xid, custName))
			return false;

		if (getLatestReservations(xid, custName).contains(rid))
			return false;

		logManager.log(xid, rid, LogType.UPDATE_RESOURCE, new ResourceData(
				data.totalAvailable, data.reserved + 1, data.price), null);
		logManager.log(xid, custName, LogType.INSERT_RESERVATION, null, rid);

		return true;

	}

	public void dieNow() {
		System.exit(1);
	}

	public void dieBeforePointerSwitch() {
		dieBeforePointerSwitch = true;
	}

	public void dieAfterPointerSwitch() {
		dieAfterPointerSwitch = true;
	}
	
	@Override
	public void dieAfterPrepare() throws RemoteException {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void dieBeforePrepare() throws RemoteException {
		// TODO Auto-generated method stub
		
	}

	private void validityCheck(int xid) throws TransactionAbortedException,
			InvalidTransactionException {
		TransactionStatus status = distributedTransactions.get(xid);
		if (status == null)
			throw new InvalidTransactionException(xid, "unknown transaction");
		if (status == TransactionStatus.aborted)
			throw new TransactionAbortedException(xid,
					"transaction already aborted");
	}

	private void updateDatabase(int xid) {
		LinkedList<LogEntry> redoOperations = logManager.getLogEntries(xid);
		if (redoOperations == null)
			return;

		LogEntry current = null;
		Iterator<LogEntry> it = redoOperations.iterator();
		while (it.hasNext()) {
			current = it.next();
			switch (current.type) {
			case INSERT_CUSTOMER:
				Set<String> customerReservations = reservations
						.get(current.resourceID);
				if (customerReservations == null)
					reservations.put(current.resourceID, new HashSet<String>());
				break;
			case INSERT_RESOURCE:
				resources.put(current.resourceID, current.afterImage);
				break;
			case UPDATE_RESOURCE:
				resources.put(current.resourceID, current.afterImage);
				break;
			case INSERT_RESERVATION:
				reservations.get(current.resourceID).add(current.reservation);
				break;
			}
		}
	}

	private ResourceData getLatestResourceData(int xid, String rid) {
		ResourceData data = logManager.insertedResource(xid, rid);
		ResourceData dataUpdate = logManager.getLatestUpdate(xid, rid);
		if (dataUpdate != null)
			data = dataUpdate;
		if (data == null)
			data = resources.get(rid);

		return data;
	}

	private Set<String> getLatestReservations(int xid, String customer) {
		Set<String> latestReservations = new HashSet<String>();
		if (reservations.containsKey(customer))
			latestReservations.addAll(reservations.get(customer));
		latestReservations.addAll(logManager.getReservations(xid, customer));

		return latestReservations;
	}

	private void readState() throws IOException {
		if (rmState.exists()) {
			DataInputStream dis = new DataInputStream(new FileInputStream(
					this.rmState));
			fileSwitch = dis.readInt();
			dis.close();
		}
	}

	private void writeState() throws IOException {
		DataOutputStream dos = new DataOutputStream(new FileOutputStream(
				rmState));
		dos.writeInt(fileSwitch);
		dos.flush();
		dos.close();
	}

	private void writeShadow() throws IOException {
		File db = fileSwitch == 0 ? db1 : db0;
		ObjectOutputStream oos = new ObjectOutputStream(
				new FileOutputStream(db));
		oos.writeObject(resources);
		oos.writeObject(reservations);
		oos.flush();
		oos.close();
	}

	private void switchPointer() throws IOException {
		fileSwitch = fileSwitch == 0 ? 1 : 0;
		writeState();
	}

	@SuppressWarnings("unchecked")
	private void readDatabase() throws IOException {
		File db = fileSwitch == 0 ? db0 : db1;
		if (db.exists()) {
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(
					db));
			try {
				resources = (Hashtable<String, ResourceData>) ois.readObject();
				reservations = (Hashtable<String, Set<String>>) ois
						.readObject();
			} catch (ClassNotFoundException e) {
				throw new IOException(e);
			}
			ois.close();
		} else {
			System.out
					.println("database file not found, creating new database");
		}
	}


}
