package transaction;

import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;

public class TravelAgencyImpl implements ITravelAgency {
	
	private IResourceManager hotels;
	private IResourceManager flights;
	private IResourceManager cars;
	
	private ITransactionManager tm;
	
	//at most one transaction must be active
	private int currentGlobalXID;
	private boolean initTxHotel = false, initTxFlight = false, initTxCar = false; 
	
	public TravelAgencyImpl() {
		try {
			Registry registry = LocateRegistry.getRegistry();
			
			this.hotels = (IResourceManager) registry.lookup(IResourceManager.RMI_NAME_HOTEL_RM);
			this.flights = (IResourceManager) registry.lookup(IResourceManager.RMI_NAME_FLIGHT_RM);
			this.cars = (IResourceManager) registry.lookup(IResourceManager.RMI_NAME_CAR_RM);
			
			this.tm = (ITransactionManager)registry.lookup(ITransactionManager.RMI_TM_NAME);
			
			currentGlobalXID = 0;
			
		} catch (AccessException e) {
			throw new RuntimeException(e);
		} catch (RemoteException e) {
			throw new RuntimeException(e);
		} catch (NotBoundException e) {
			throw new RuntimeException(e);
		}
	}
	
	//******************* TRANSACTION *********************

	@Override
	public boolean abort() {
		if(validTransaction())
			try {
				boolean success = tm.abort(currentGlobalXID);
				if(success){
					currentGlobalXID = 0;
					initTxHotel = initTxFlight = initTxCar = false;
				}
				return success;
			} catch (RemoteException e) {
				e.printStackTrace();
				return false;
			} catch (TransactionAbortedException e) {
				e.printStackTrace();
				return false;
			} catch (InvalidTransactionException e) {
				e.printStackTrace();
				return false;
			}
		else
			return false;
	}

	@Override
	public boolean commit() {
		if(validTransaction())
			try {
				boolean success = tm.commit(currentGlobalXID);
				if(success){
					currentGlobalXID = 0;
					initTxHotel = initTxFlight = initTxCar = false;
				}
				return success;
			} catch (RemoteException e) {
				e.printStackTrace();
				return false;
			} catch (TransactionAbortedException e) {
				e.printStackTrace();
				return false;
			} catch (InvalidTransactionException e) {
				e.printStackTrace();
				return false;
			}
		else
			return false;
	}
	
	//******************* RESERVATION *********************

	@Override
	public boolean reserve(ResourceType resourceType, String rid, String customerName) {
		switch(resourceType){
		case HOTEL:
			return reserve(hotels, rid, customerName);
		case FLIGHT:
			return reserve(flights, rid, customerName);
		case CAR:
			return reserve(cars, rid, customerName);
		}
		
		return false;
	}	

	@Override
	public boolean reserveItinerary(String hotelName, String flightNumber,
			String agencyName, String customerName) {
		
		boolean success = reserve(cars, agencyName, customerName);
		success &= reserve(flights, flightNumber, customerName);
		success &= reserve(hotels, hotelName, customerName);
		
		return success;
	}
	
	
	//******************* QUERY *********************
	
	@Override
	public int queryAvailability(ResourceType resourceType, String rid) {	
		try {
			switch(resourceType){
			case HOTEL:
				checkTransactionStart(hotels);
				return hotels.queryResourceAvailability(currentGlobalXID, rid);
			case FLIGHT:
				checkTransactionStart(flights);
				return flights.queryResourceAvailability(currentGlobalXID, rid);
			case CAR:
				checkTransactionStart(cars);
				return cars.queryResourceAvailability(currentGlobalXID, rid);
			}
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		}
		
		return -1;	
	}	
	
	@Override
	public int queryPrice(ResourceType resourceType, String rid) {	
		try {
			switch(resourceType){
			case HOTEL:
				checkTransactionStart(hotels);
				return hotels.queryResourcePrice(currentGlobalXID, rid);
			case FLIGHT:
				checkTransactionStart(flights);
				return flights.queryResourcePrice(currentGlobalXID, rid);
			case CAR:
				checkTransactionStart(cars);
				return cars.queryResourcePrice(currentGlobalXID, rid);
			}
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		}
		
		return -1;		
	}	
	
	@Override
	public int queryBill(String customerName) {
		int billHotel, billFlight, billCar;
		
		billHotel = queryBill(hotels, customerName);
		billFlight = queryBill(flights, customerName);
		billCar = queryBill(cars, customerName);
		
		return billHotel >= 0 ? billHotel : 0 +
				billFlight >= 0 ? billFlight : 0 +
				billCar >= 0 ? billCar : 0;  
					
	}
	
	//******************* ADMINISTRATION INTERFACE *********************
	
	
	

	
	@Override
	public boolean addResource(ResourceType resourceType, String rid,
			int num, int price) {
		try {
			switch(resourceType){
			case HOTEL:
				checkTransactionStart(hotels);
				return hotels.addResource(currentGlobalXID, rid, num, price);
			case FLIGHT:
				checkTransactionStart(flights);
				return flights.addResource(currentGlobalXID, rid, num, price);
			case CAR:
				checkTransactionStart(cars);
				return cars.addResource(currentGlobalXID, rid, num, price);
			}
		} catch (RemoteException e) {
			e.printStackTrace();			
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		}
		
		return false;
	}

	@Override
	public boolean newCustomer(ResourceType resourceType, String custName) {
		
		try {
			switch(resourceType){
			case HOTEL:
				checkTransactionStart(hotels);
				return hotels.newCustomer(currentGlobalXID, custName);
			case FLIGHT:
				checkTransactionStart(flights);
				return flights.newCustomer(currentGlobalXID, custName);
			case CAR:
				checkTransactionStart(cars);
				return cars.newCustomer(currentGlobalXID, custName);
			}
		} catch (RemoteException e) {
			e.printStackTrace();			
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		}
		
		return false;
	}

	@Override
	public boolean updateResource(ResourceType resourceType, String rid, int num, int price) {
		try {
			switch(resourceType){
			case HOTEL:
				checkTransactionStart(hotels);
				return hotels.updateResource(currentGlobalXID, rid, num, price);
			case FLIGHT:
				checkTransactionStart(flights);
				return flights.updateResource(currentGlobalXID, rid, num, price);
			case CAR:
				checkTransactionStart(cars);
				return cars.updateResource(currentGlobalXID, rid, num, price);
			}
		} catch (RemoteException e) {
			e.printStackTrace();			
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		}
		
		return false;
	}
	
	
	@Override
	public void dieAfterCollected() throws RemoteException {		
		tm.dieAfterCollected();
	}

	@Override
	public void dieAfterPrepare(ResourceType type) throws RemoteException {
		switch(type){
		case HOTEL:
			hotels.dieAfterPrepare();
			break;
		case FLIGHT:
			flights.dieAfterPrepare();
			break;
		case CAR:
			cars.dieAfterPrepare();
			break;
		}
	}

	@Override
	public void dieBeforeCollected() throws RemoteException {
		tm.dieBeforeCollected();
	}

	@Override
	public void dieBeforePrepare(ResourceType type) throws RemoteException {
		switch(type){
		case HOTEL:
			hotels.dieBeforePrepare();
			break;
		case FLIGHT:
			flights.dieBeforePrepare();
			break;
		case CAR:
			cars.dieBeforePrepare();
			break;
		}		
	}
	
	//******************* CONVENIENCE METHODS *********************

	private boolean reserve(IResourceManager rm, String rid, String customer){
		try {
			checkTransactionStart(rm);
			int available = rm.queryResourceAvailability(currentGlobalXID, rid);
			if(available > 0){
				boolean success = rm.reserveResource(currentGlobalXID, rid, customer);
				if(!success){
					rm.newCustomer(currentGlobalXID, customer);
					success = rm.reserveResource(currentGlobalXID, rid, customer);
				}
				
				return success;
			}
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		}
		
		return false;
	}
	
	private int queryBill(IResourceManager rm, String customer){
		try {
			checkTransactionStart(rm);
			return rm.queryCustomerBill(currentGlobalXID, customer);
		} catch (RemoteException e) {			
			e.printStackTrace();
		} catch (ShutdownException e) {
			e.printStackTrace();
		} catch (TransactionAbortedException e) {
			e.printStackTrace();
		} catch (InvalidTransactionException e) {
			e.printStackTrace();
		}
		
		return -1;
	}
	
	private boolean validTransaction(){
		return currentGlobalXID != 0;
	}
	
	private void checkTransactionStart(IResourceManager rm)
	throws RemoteException, ShutdownException, InvalidTransactionException{
		if(currentGlobalXID == 0){
			currentGlobalXID = tm.begin();
		}
		
		if(rm == hotels){
			if(!initTxHotel){
				hotels.beginDistributed(currentGlobalXID);
				initTxHotel = true;
			}
		}
		else if(rm == flights){
			if(!initTxFlight){
				flights.beginDistributed(currentGlobalXID);
				initTxFlight = true;
			}
		}
		else if(rm == cars){
			if(!initTxCar){
				cars.beginDistributed(currentGlobalXID);
				initTxCar = true;
			}
		}
	}
}
